/* 
 * LIFE 3d vedi le scienze Aprile 1987 numero 224 Scientific American 
 * pag. 98 {ri} Creazioni al calcolatore 
 *
 * thanks to
 *    Example program for the Allegro library, by Shawn Hargreaves.
 * bouncing cubes
 */


#include <stdlib.h>
#include <stdio.h>

#include "allegro.h"


#define MAX_SHAPES         1000  /* number of bouncing cubes */

#define NUM_VERTICES       8     /* a cube has eight corners */
#define NUM_FACES          6     /* a cube has six faces */

int  	cellule[2][10][10][10],
		stato,
		quale,adiac,fi,fs,ai,as;

int num_shapes;

typedef struct VTX		 /* vertice   */
{
   fixed x, y, z;
} VTX;


typedef struct QUAD              /* four vertices makes a quad */
{
   VTX *vtxlist;
   int v1, v2, v3, v4;
} QUAD;


typedef struct SHAPE             /* store position of a shape */
{
   fixed x, y, z;                /* x, y, z position */
   fixed rx, ry, rz;             /* rotations */
   fixed dz;                     /* speed of movement */
   fixed drx, dry, drz;          /* speed of rotation */
} SHAPE;


VTX points[] =                   /* a cube, centered on the origin */
{
   /* vertices of the cube */
   { -32 << 16, -32 << 16, -32 << 16 },
   { -32 << 16,  32 << 16, -32 << 16 },
   {  32 << 16,  32 << 16, -32 << 16 },
   {  32 << 16, -32 << 16, -32 << 16 },
   { -32 << 16, -32 << 16,  32 << 16 },
   { -32 << 16,  32 << 16,  32 << 16 },
   {  32 << 16,  32 << 16,  32 << 16 },
   {  32 << 16, -32 << 16,  32 << 16 },
};


QUAD faces[] =                   /* group the vertices into polygons */
{
   { points, 0, 3, 2, 1 },
   { points, 4, 5, 6, 7 },
   { points, 0, 1, 5, 4 },
   { points, 2, 3, 7, 6 },
   { points, 0, 4, 7, 3 },
   { points, 1, 2, 6, 5 }
};


SHAPE shapes[MAX_SHAPES];        /* a list of shapes */


/* somewhere to put translated vertices */
VTX output_points[NUM_VERTICES * MAX_SHAPES];
QUAD output_faces[NUM_FACES * MAX_SHAPES];


enum { 
   wireframe,
   flat,
   gcol,
   grgb,
   atex,
   ptex,
   atex_mask,
   ptex_mask,
   atex_lit,
   ptex_lit,
   atex_mask_lit,
   ptex_mask_lit,
   last_mode
} render_mode = wireframe;


int render_type[] = {
   0,
   POLYTYPE_FLAT,
   POLYTYPE_GCOL,
   POLYTYPE_GRGB,
   POLYTYPE_ATEX,
   POLYTYPE_PTEX,
   POLYTYPE_ATEX_MASK,
   POLYTYPE_PTEX_MASK,
   POLYTYPE_ATEX_LIT,
   POLYTYPE_PTEX_LIT,
   POLYTYPE_ATEX_MASK_LIT,
   POLYTYPE_PTEX_MASK_LIT
};

char *mode_desc[] = {
   "Wireframe",
   "Flat shaded",
   "Single color gouraud shaded",
   "Gouraud shaded",
   "Texture mapped",
   "Perspective correct texture mapped",
   "Masked texture mapped",
   "Masked persp. correct texture mapped",
   "Lit texture map",
   "Lit persp. correct texture map",
   "Masked lit texture map",
   "Masked lit persp. correct texture map"
};


BITMAP *texture;



/* initialise shape positions */
void init_shapes()
{
   int c;

   for (c=0; c<num_shapes; c++) {
      shapes[c].x = itofix(0);
      shapes[c].y = itofix(0);
      shapes[c].z = itofix(0);
      shapes[c].rx = 0;
      shapes[c].ry = 0;
      shapes[c].rz = 0;
      shapes[c].dz =  0;
      shapes[c].drx = 0;
      shapes[c].dry = 0;
      shapes[c].drz = 0;
   }
}


/* update shape positions */
void animate_shapes()
{
   int c;

   for (c=0; c<num_shapes; c++) {
      shapes[c].z += shapes[c].dz;

      if ((shapes[c].z > itofix(1024)) ||
	  (shapes[c].z < itofix(192)))
	 shapes[c].dz = -shapes[c].dz;

      shapes[c].rx += shapes[c].drx;
      shapes[c].ry += shapes[c].dry;
      shapes[c].rz += shapes[c].drz;
   }
}


/* translate shapes from 3d world space to 2d screen space */
void translate_shapes()
{
   int c, d;
   MATRIX matrix;
   VTX *outpoint = output_points;
   QUAD *outface = output_faces;

   for (c=0; c<num_shapes; c++) {
      /* build a transformation matrix */
      get_transformation_matrix(&matrix, itofix(1),
				shapes[c].rx, shapes[c].ry, shapes[c].rz,
				shapes[c].x, shapes[c].y, shapes[c].z);

      /* output the vertices */
      for (d=0; d<NUM_VERTICES; d++) {
	 apply_matrix(&matrix, points[d].x, points[d].y, points[d].z, &outpoint[d].x, &outpoint[d].y, &outpoint[d].z);
	 persp_project(outpoint[d].x, outpoint[d].y, outpoint[d].z, &outpoint[d].x, &outpoint[d].y);
      }

      /* output the faces */
      for (d=0; d<NUM_FACES; d++) {
	 outface[d] = faces[d];
	 outface[d].vtxlist = outpoint;
      }

      outpoint += NUM_VERTICES;
      outface += NUM_FACES;
   }
}


/* draw a line (for wireframe display) */
void wire(BITMAP *b, VTX *v1, VTX *v2)
{
   int col = MID(128, 255 - fixtoi(v1->z+v2->z) / 16, 255);
   line(b, fixtoi(v1->x), fixtoi(v1->y), fixtoi(v2->x), fixtoi(v2->y), palette_color[col]);
}


/* draw a quad */
void quad(BITMAP *b, VTX *v1, VTX *v2, VTX *v3, VTX *v4, int mode)
{
   int col;

   /* four vertices */
   V3D vtx1 = { v1->x, v1->y, v1->z, 0,      0,      0 };
   V3D vtx2 = { v2->x, v2->y, v2->z, 31<<16, 0,      0 };
   V3D vtx3 = { v3->x, v3->y, v3->z, 31<<16, 31<<16, 0 };
   V3D vtx4 = { v4->x, v4->y, v4->z, 0,      31<<16, 0 };

   /* cull backfaces */
   if ((mode != POLYTYPE_ATEX_MASK) && (mode != POLYTYPE_PTEX_MASK) &&
       (mode != POLYTYPE_ATEX_MASK_LIT) && (mode != POLYTYPE_PTEX_MASK_LIT) &&
       (polygon_z_normal(&vtx1, &vtx2, &vtx3) < 0))
      return;

   /* set up the vertex color, differently for each rendering mode */
   switch (mode) {

      case POLYTYPE_FLAT:
	 col = MID(128, 255 - fixtoi(v1->z+v2->z) / 16, 255);
	 vtx1.c = vtx2.c = vtx3.c = vtx4.c = palette_color[col];
	 break;

      case POLYTYPE_GCOL:
	 vtx1.c = palette_color[0xD0];
	 vtx2.c = palette_color[0x80];
	 vtx3.c = palette_color[0xB0];
	 vtx4.c = palette_color[0xFF];
	 break;

      case POLYTYPE_GRGB:
	 vtx1.c = 0x000000;
	 vtx2.c = 0x7F0000;
	 vtx3.c = 0xFF0000;
	 vtx4.c = 0x7F0000;
	 break;

      case POLYTYPE_ATEX_LIT:
      case POLYTYPE_PTEX_LIT:
      case POLYTYPE_ATEX_MASK_LIT:
      case POLYTYPE_PTEX_MASK_LIT:
	 vtx1.c = MID(0, 255 - fixtoi(v1->z) / 4, 255);
	 vtx2.c = MID(0, 255 - fixtoi(v2->z) / 4, 255);
	 vtx3.c = MID(0, 255 - fixtoi(v3->z) / 4, 255);
	 vtx4.c = MID(0, 255 - fixtoi(v4->z) / 4, 255);
	 break; 
   }

   /* draw the quad */
   quad3d(b, mode, texture, &vtx1, &vtx2, &vtx3, &vtx4);
}


/* callback for qsort() */
int quad_cmp(const void *e1, const void *e2)
{
   QUAD *q1 = (QUAD *)e1;
   QUAD *q2 = (QUAD *)e2;

   fixed d1 = q1->vtxlist[q1->v1].z + q1->vtxlist[q1->v2].z +
	      q1->vtxlist[q1->v3].z + q1->vtxlist[q1->v4].z;

   fixed d2 = q2->vtxlist[q2->v1].z + q2->vtxlist[q2->v2].z +
	      q2->vtxlist[q2->v3].z + q2->vtxlist[q2->v4].z;

   return d2 - d1;
}


/* draw the shapes calculated by translate_shapes() */
void draw_shapes(BITMAP *b)
{
   int c;
   QUAD *face = output_faces;
   VTX *v1, *v2, *v3, *v4;

   /* depth sort */
   qsort(output_faces, NUM_FACES * num_shapes, sizeof(QUAD), quad_cmp);

   for (c=0; c < NUM_FACES * num_shapes; c++) {
      /* find the vertices used by the face */
      v1 = face->vtxlist + face->v1;
      v2 = face->vtxlist + face->v2;
      v3 = face->vtxlist + face->v3;
      v4 = face->vtxlist + face->v4;

      /* draw the face */
      if (render_mode == wireframe) {
	 wire(b, v1, v2);
	 wire(b, v2, v3);
	 wire(b, v3, v4);
	 wire(b, v4, v1);
      }
      else {
	 quad(b, v1, v2, v3, v4, render_type[render_mode]);
      }

      face++;
   }
}


/* RGB -> color mapping table. Not needed, but speeds things up */
RGB_MAP rgb_table;

/* lighting color mapping table */
COLOR_MAP light_table;


void print_progress(int pos)
{
   if ((pos & 3) == 3) {
      printf("*");
      fflush(stdout);
   }
}



void adiacenti(int i,int j,int k, int *tot)
{
 int l,m,n,who;

 *tot = 0;
 who = !quale;
 for (l = i-1; l <= i+1; l++)
	for (m = j-1 ; m <= j+1; m++)
		for (n = k-1; n <= k+1; n++)
			if (cellule[who][l][m][n])
			    (*tot)++;
 if (cellule[who][i][j][k]) (*tot)--;
}

void chestato(int i, int j, int k, int quanti)
{
 int other;

 other = !quale;
 if (cellule[other][i][j][k])
 {
  if ((quanti<ai) || (quanti>as))
	  cellule[quale][i][j][k] = 0;
	else
	  cellule[quale][i][j][k] = 1;
 }
 else
 {
  if ((quanti<fi) || (quanti>fs))
          cellule[quale][i][j][k] = 0;
        else
          cellule[quale][i][j][k] = 1;
 }
}

void eraseall(void)
{
 int i,x,y,z;

 for (z=9; z>=0; z--)
	for (y=9; y>=0; y--)
		for (x=9; x>=0; x--)
			for (i=0; i<2; i++)
				cellule[i][x][y][z] = 0;
}

void random_fill(void)
{
 int x,y,z;

 for (z=9; z>=0; z--)
  for (y=9; y>=0; y--)
   for (x=9; x>=0; x--)
    cellule[1][x][y][z] = 1; //rand()%2 ;

}

void calc_shapes() {
 int x,y,z;

 num_shapes = 0;

 for (z=9; z>=0; z--)
  for (y=9; y>=0; y--)
   for (x=9; x>=0; x--)
    { 
      if (cellule[quale][x][y][z]) 
      {
       shapes[num_shapes].x = itofix(65*x-300);
       shapes[num_shapes].y = itofix(65*y-300);
       shapes[num_shapes].z = itofix(65*z+400);
        
       ++num_shapes;
      }
    }
}

/*
void input(void)
{
 z ==8; x =8; y=8;
 do
  {
	a=600;b =0; c=-100;
	x1=x*65+a; y1 = y*65+b; z1 = z*65;
	calc2d(x1,y1,z1,1);
	wires(x1,y1,z1,1);
	car = getch();
	switch car
	{
	  case 'y': cellule[1][x][y][z] =1;
					drawcube(x1,y1,z1);
					break
	  case '6': if x<8 x++;
					break;
	  case '4': if x>1 x--;
					break;
	  case '8': if y>1 y--;
	  case '2': if y<8 y++;
	  case '5': z--;
	  case 27 : return;
	}
   wires(x1,y1,z1,0);
  }
  while ((car = 27) or (z=0))
}
*/

int main(void)
{
 BITMAP *buffer;
   PALLETE pal;
   int c = GFX_AUTODETECT;
   int w = 320;
   int h = 200;
   int bpp = 8;
   int last_retrace_count;
   int x,y,z;

   render_mode = 1; // modalita di renderizzazione


   // init all************************************************************
   allegro_init();
   install_keyboard();
   install_mouse();
   install_timer();
   
   /* color 0 = black */
   pal[0].r = pal[0].g = pal[0].b = 0;

   /* copy the desktop pallete */
   for (c=1; c<64; c++)
      pal[c] = desktop_pallete[c];

   /* make a red gradient */
   for (c=64; c<96; c++) {
      pal[c].r = (c-64)*2;
      pal[c].g = pal[c].b = 0;
   }

   /* make a green gradient */
   for (c=96; c<128; c++) {
      pal[c].g = (c-96)*2;
      pal[c].r = pal[c].b = 0;
   }

   /* set up a greyscale in the top half of the pallete */
   for (c=128; c<256; c++)
      pal[c].r = pal[c].g = pal[c].b = (c-128)/2;

   /* set the graphics mode */
   set_gfx_mode(GFX_VGA, 320, 200, 0, 0);
   set_pallete(desktop_pallete);

   if (!gfx_mode_select_ex(&c, &w, &h, &bpp)) {
      allegro_exit();
      return 1;
   }

   set_color_depth(bpp);

   if (set_gfx_mode(c, w, h, 0, 0) != 0) {
      allegro_exit();
      printf("Error setting graphics mode\n%s\n\n", allegro_error);
      return 1;
   }

   if (gfx_driver->id == GFX_MODEX) {
      allegro_exit();
      printf("Can't draw 3d polygons in mode-X, sorry!\n\n");
      return 1;
   }

   set_pallete(pal);
   /* double buffer the animation */
   buffer = create_bitmap(SCREEN_W, SCREEN_H);

   /* set up the viewport for the perspective projection */
   set_projection_viewport(0, 0, SCREEN_W, SCREEN_H);

// init all************************************************************
   
   // valori di nascita/morte delle cellule
   ai = 4; as =5; fi =5; fs =5;

   eraseall(); // cancella caselle nella struttura
   //input();
   random_fill(); // riempie matrice a caso

   /* initialise the cubes */
   /* calcola posizioni cubi */
   init_shapes();

   quale = 0;
   do
   {
    clear(buffer);
    for (z=8; z>0; z--)
	for (y=8; y>0; y--)
	 for (x=8; x>0; x--)
	 {
	  adiacenti(x,y,z,&adiac);
	  chestato(x,y,z,adiac);
	 }
    //calcola nuove posizioni cubi
    calc_shapes();

    quale = !quale;

    // calcola proiezione e stampa
    translate_shapes();
    draw_shapes(buffer);
    vsync();
    blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
  
    // stoppa se esc
    if (keypressed()) 
	 if ((readkey() & 0xFF) == 27)
	    break;
   }
   while (1) ;

   destroy_bitmap(buffer);

   return 0;
}

